import torch, pathlib
from typing import Dict, Any, Optional, Tuple
from transformers import AutoModelForCausalLM, AutoTokenizer, AutoConfig, AutoModel
from peft.tuners.lora import LoraLayer
from peft import LoraConfig, get_peft_model, set_peft_model_state_dict
try:                from torch.utils.checkpoint import CheckpointImpl
except ImportError: from torch.distributed.algorithms._checkpoint.checkpoint_wrapper import CheckpointImpl
from torch.distributed.algorithms._checkpoint.checkpoint_wrapper import checkpoint_wrapper, apply_activation_checkpointing

def build_critic_cls(base_cls, base_pretrain_cls, value_head_prefix):
    class CriticModel(base_pretrain_cls):
        supports_gradient_checkpointing = True
        def __init__(self, config: AutoConfig):
            super().__init__(config)
            setattr(self, self.base_model_prefix, base_cls(config))
            self.value_head_prefix = value_head_prefix
            setattr(self, value_head_prefix, torch.nn.Linear(config.hidden_size, 1, bias=False))

        def forward(self, input_ids: torch.LongTensor=None, attention_mask: Optional[torch.Tensor]=None, return_output=False, **_) -> torch.Tensor:
            position_ids = attention_mask.long().cumsum(-1) - 1; position_ids.masked_fill_(attention_mask == 0, 1)
            outputs = getattr(self, self.base_model_prefix)(input_ids, attention_mask=attention_mask, position_ids=position_ids)
            last_hidden_states = outputs["last_hidden_state"]; values = getattr(self, self.value_head_prefix)(last_hidden_states).squeeze(-1)
            if return_output:   return (values, outputs)
            else:               return values
    return CriticModel

def get_critic_model(pretrain_or_model: str, device: torch.device, torch_dtype, use_flash_attention_2: bool=False, value_head_prefix: str="value_head"):
    config = AutoConfig.from_pretrained(pretrain_or_model, trust_remote_code=True)
    config._attn_implementation = "flash_attention_2" if use_flash_attention_2 else "eager"
    base_class = AutoModel._model_mapping[type(config)]
    critic_cls = build_critic_cls(base_class, base_class.__base__, value_head_prefix)
    model = critic_cls.from_pretrained(pretrain_or_model, config=config, trust_remote_code=True, torch_dtype=torch_dtype, device_map=device)
    value_head = getattr(model, value_head_prefix)
    value_head.weight.data.normal_(mean=0.0, std=1 / (config.hidden_size + 1))
    return model

def _load_base(name: str, dtype, device, **kwargs): 
    with torch.device(device): 
        return AutoModelForCausalLM.from_pretrained(name, torch_dtype=dtype, trust_remote_code=True, **kwargs)

def _freeze(model, ignore_substr: Optional[str] = None):
    for n, p in model.named_parameters():
        if ignore_substr and ignore_substr in n: continue
        p.requires_grad_(False)

def _build_lora(model, lora_cfg: Dict[str, Any], task_type: str):
    return get_peft_model(model, LoraConfig(
        r=lora_cfg.get("lora_rank", 32), lora_alpha=lora_cfg.get("lora_alpha", 32), lora_dropout=lora_cfg.get("lora_dropout", 0.05),
        bias="none", task_type=task_type, target_modules=lora_cfg.get("target_modules", ["q_proj", "k_proj", "v_proj", "o_proj"]),
    ))

def build_peft_model(base_name: str, device: torch.device, lora_cfg: Dict[str, Any]|None, initial_lora_path: Optional[str]=None, freeze_base: bool=True, critic_model: bool=False, value_head_prefix: str="value_head") -> Tuple[torch.nn.Module, "transformers.PreTrainedTokenizer"]:
    task_type = "TOKEN_CLS" if critic_model else "CAUSAL_LM"
    base = get_critic_model(base_name, device, torch_dtype=torch.bfloat16, value_head_prefix=value_head_prefix) if critic_model else _load_base(base_name, torch.bfloat16, device)
    if freeze_base: _freeze(base, None if not critic_model else value_head_prefix)
    model = _build_lora(base, lora_cfg or {}, task_type).to(device)
    if initial_lora_path: _load_lora_state(model, initial_lora_path)
    tok = AutoTokenizer.from_pretrained(base_name, trust_remote_code=True)
    tok.pad_token = tok.eos_token
    return model, tok

def _json_safe(obj):
    if isinstance(obj, set): return list(obj) # turn sets into lists
    raise TypeError # let json handle the rest
    
def enable_full_activation_ckpt(model):
    def checkpoint_everything(mod):
        if isinstance(mod, LoraLayer): return False
        for _, child in mod.named_modules():
            if isinstance(child, LoraLayer): return False
        return True
    apply_activation_checkpointing(model, checkpoint_wrapper_fn=lambda m: checkpoint_wrapper(m, checkpoint_impl=CheckpointImpl.NO_REENTRANT), check_fn=checkpoint_everything)  # "always recompute"



import os, ray, tree, random
from threading import Lock
from typing import List, Dict, Optional, Tuple, Callable

# local imports
from unstable.utils.logging import setup_logger
from unstable._types import PlayerTrajectory, Step
# from unstable.core import BaseTracker
from unstable.trackers import BaseTracker
from unstable.utils import write_training_data_to_file
from unstable.reward_transformations import ComposeFinalRewardTransforms, ComposeStepRewardTransforms, ComposeSamplingRewardTransforms


class BaseBuffer:
    def __init__(self, max_buffer_size: int, tracker: BaseTracker, final_reward_transformation: Optional[ComposeFinalRewardTransforms], step_reward_transformation: Optional[ComposeStepRewardTransforms], sampling_reward_transformation: Optional[ComposeSamplingRewardTransforms], buffer_strategy: str = "random"): ...
    def add_player_trajectory(self, player_traj: PlayerTrajectory, env_id: str): ...
    def get_batch(self, batch_size: int): ...



@ray.remote
class StepBuffer(BaseBuffer):
    def __init__(
        self, max_buffer_size: int, tracker: BaseTracker, 
        final_reward_transformation: Optional[ComposeFinalRewardTransforms], 
        step_reward_transformation: Optional[ComposeStepRewardTransforms], 
        sampling_reward_transformation: Optional[ComposeSamplingRewardTransforms], 
        buffer_strategy: str = "random"
    ):
        self.max_buffer_size, self.buffer_strategy = max_buffer_size, buffer_strategy
        self.final_reward_transformation = final_reward_transformation
        self.step_reward_transformation = step_reward_transformation
        self.sampling_reward_transformation = sampling_reward_transformation
        self.collect = True
        self.steps: List[Step] = []
        self.training_steps = 0
        self.tracker = tracker
        self.local_storage_dir = ray.get(self.tracker.get_train_dir.remote())
        self.logger = setup_logger("step_buffer", ray.get(tracker.get_log_dir.remote())) # setup logging
        self.mutex = Lock()

    def add_player_trajectory(self, player_traj: PlayerTrajectory, env_id: str):
        reward = self.final_reward_transformation(reward=player_traj.final_reward, pid=player_traj.pid, env_id=env_id) if self.final_reward_transformation else player_traj.final_reward
        for idx in range(len(player_traj.obs)):
            step_reward = self.step_reward_transformation(player_traj=player_traj, step_index=idx, reward=reward) if self.step_reward_transformation else reward
            with self.mutex: 
                self.steps.append(Step(pid=player_traj.pid, obs=player_traj.obs[idx], act=player_traj.actions[idx], reward=step_reward, env_id=env_id, step_info={"raw_reward": player_traj.final_reward, "env_reward": reward, "step_reward": step_reward}))
        self.logger.info(f"Buffer size: {len(self.steps)}, added {len(player_traj.obs)} steps")
        # downsample if necessary
        excess_num_samples = max(0, len(self.steps) - self.max_buffer_size)
        self.logger.info(f"Excess Num Samples: {excess_num_samples}")
        if excess_num_samples > 0:
            self.logger.info(f"Downsampling buffer because of excess samples")
            with self.mutex: 
                randm_sampled = random.sample(self.steps, excess_num_samples)
                for b in randm_sampled:
                    self.steps.remove(b)
                self.logger.info(f"Buffer size after downsampling: {len(self.steps)}")

    def get_batch(self, batch_size: int) -> List[Step]:
        with self.mutex: 
            batch = random.sample(self.steps, batch_size)
            for b in batch: self.steps.remove(b)
        batch = self.sampling_reward_transformation(batch) if self.sampling_reward_transformation is not None else batch
        self.logger.info(f"Sampling {len(batch)} samples from buffer.")
        try: write_training_data_to_file(batch=batch, filename=os.path.join(self.local_storage_dir, f"train_data_step_{self.training_steps}.csv"))
        except Exception as exc: self.logger.error(f"Exception when trying to write training data to file: {exc}")
        self.training_steps += 1
        return batch

    def stop(self):                 self.collect = False
    def size(self) -> int:          return len(self.steps)
    def continue_collection(self):  return self.collect
    def clear(self):                
        with self.mutex: 
            self.steps.clear()


@ray.remote
class EpisodeBuffer(BaseBuffer):
    def __init__(
        self, max_buffer_size: int, tracker: BaseTracker, 
        final_reward_transformation: Optional[ComposeFinalRewardTransforms], 
        step_reward_transformation: Optional[ComposeStepRewardTransforms], 
        sampling_reward_transformation: Optional[ComposeSamplingRewardTransforms], 
        buffer_strategy: str = "random"
    ):
        self.max_buffer_size, self.buffer_strategy = max_buffer_size, buffer_strategy
        self.final_reward_transformation = final_reward_transformation
        self.step_reward_transformation = step_reward_transformation
        self.sampling_reward_transformation = sampling_reward_transformation
        self.collect = True
        self.training_steps = 0
        self.tracker = tracker
        self.local_storage_dir = ray.get(self.tracker.get_train_dir.remote())
        self.logger = setup_logger("step_buffer", ray.get(tracker.get_log_dir.remote()))  # setup logging
        self.episodes: List[List[Step]] = []
        self.mutex = Lock()

    def add_player_trajectory(self, player_traj: PlayerTrajectory, env_id: str):
        episode = []
        reward = self.final_reward_transformation(reward=player_traj.final_reward, pid=player_traj.pid, env_id=env_id) if self.final_reward_transformation else player_traj.final_reward
        for idx in range(len(player_traj.obs)):
            step_reward = self.step_reward_transformation(player_traj=player_traj, step_index=idx, reward=reward) if self.step_reward_transformation else reward
            episode.append(Step(pid=player_traj.pid, obs=player_traj.obs[idx], act=player_traj.actions[idx], reward=step_reward, env_id=env_id, step_info={"raw_reward": player_traj.final_reward, "env_reward": reward, "step_reward": step_reward}))
        with self.mutex:
            self.episodes.append(episode)
            excess_num_samples = max(0, len(tree.flatten(self.episodes)) - self.max_buffer_size)
            self.logger.info(f"BUFFER NUM of STEP {len(tree.flatten(self.episodes))}")
            while excess_num_samples > 0:
                randm_sampled = random.sample(self.episodes, 1)
                for b in randm_sampled: self.episodes.remove(b)
                excess_num_samples = max(0, len(tree.flatten(self.episodes)) - self.max_buffer_size)
        
    def get_batch(self, batch_size: int) -> List[List[Step]]:
        with self.mutex:
            assert len(tree.flatten(self.episodes)) >= batch_size
            step_count = 0
            sampled_episodes = []
            random.shuffle(self.episodes)
            for ep in self.episodes:
                sampled_episodes.append(ep)
                step_count += len(ep)
                if step_count >= batch_size: break
            for ep in sampled_episodes: self.episodes.remove(ep)
        self.logger.info(f"Sampling {len(sampled_episodes)} episodes from buffer.")
        self.training_steps += 1
        return sampled_episodes

    def stop(self):                 self.collect = False
    def size(self) -> int:          return len(tree.flatten(self.episodes))
    def continue_collection(self):  return self.collect
    def clear(self):
        with self.mutex: 
            self.episodes.clear()



from unstable.samplers.model_samplers import BaseModelSampler, FixedOpponentModelSampler
from unstable.samplers.env_samplers import BaseEnvSampler, UniformRandomEnvSampler

__all__ = [
    "BaseModelSampler", "FixedOpponentModelSampler",
    "BaseEnvSampler", "UniformRandomEnvSampler"
]


import re, random, logging, itertools
from pathlib import Path
from dataclasses import asdict, dataclass, field
from typing import Any, Callable, Dict, List, Tuple, Protocol, Optional

import ray
from ray.exceptions import RayActorError, RayTaskError

import textarena as ta
assert ta.__version__ >= "0.6.16", f"TextArena package version is too old: {ta.__version__}. Required version is at least 0.6.16."

# local imports
from unstable.actor import VLLMActor
from unstable._types import GameSpec, GameInformation, PlayerTrajectory, TaskMeta
from unstable.utils.logging import setup_logger
from unstable.utils.templates import ACTION_EXTRACTION, OBSERVATION_FORMATTING



class CallableActorWrapper:
    def __init__(self, actor: VLLMActor, lora_path: str|Path, obs_fmt_fn: Callable[[str],str], extract_fn: Callable[[str], Tuple[str, Dict[str, Any]]]) -> None:
        self._actor, self._lora, self._fmt, self._extract = actor, lora_path, obs_fmt_fn, extract_fn

    def __call__(self, observation: str) -> str: 
        _, extracted, _, _ = self.act_full(observation)
        return extracted

    def act_full(self, observation: str) -> Tuple[str, str, str, dict]:
        prompt = self._fmt(observation=observation)
        raw = ray.get(self._actor.submit_prompt.remote(prompt=prompt, lora_path=self._lora))
        extracted, format_feedback = self._extract(raw_action=raw)
        return raw, extracted, prompt, format_feedback

@ray.remote(num_cpus=0)
def run_game(game_spec: GameSpec, actor: VLLMActor):
    game_information = GameInformation(game_idx=game_spec.game_idx, eval_model_pid=game_spec.eval_model_pid, eval_opponent_name=game_spec.eval_opponent_name)
    agents = {agent_spec.pid: {
        "traj": PlayerTrajectory(pid=agent_spec.pid) if agent_spec.collect_data else None, 
        "name": agent_spec.lora_path if agent_spec.lora_path else agent_spec.openrouter_name,
        "model": CallableActorWrapper(actor=actor, lora_path=agent_spec.lora_path, obs_fmt_fn=OBSERVATION_FORMATTING[agent_spec.prompt_template], extract_fn=ACTION_EXTRACTION[agent_spec.action_extraction_fn]) if agent_spec.openrouter_name==None else ta.agents.OpenRouterAgent(agent_spec.openrouter_name)
    } for agent_spec in game_spec.agent_specs} # build agents
    env=ta.make(game_spec.env_id); env.reset(num_players=len(agents), seed=game_spec.seed); env.state.error_allowance=0; turn=0
    while True:
        pid, obs = env.get_observation()
        # get model (or opponent) action
        if agents[pid]["traj"] == None: raw = extracted = agents[pid]["model"](obs) # fix opponent
        else: raw, extracted, prompt, format_feedback = agents[pid]["model"].act_full(obs)
        done, step_info = env.step(extracted); turn+= 1 # execute the action & increment turn counter
        # general tracking
        game_information.pid.append(pid); game_information.obs.append(obs); game_information.full_actions.append(raw)
        game_information.extracted_actions.append(extracted); game_information.step_infos.append(step_info); game_information.names[pid] = agents[pid]["name"]
        # player specific trackering
        if agents[pid]["traj"] != None:
            agents[pid]["traj"].obs.append(obs); agents[pid]["traj"].actions.append(raw); agents[pid]["traj"].extracted_actions.append(extracted)
            format_feedback["invalid_move"] = False; agents[pid]["traj"].format_feedbacks.append(format_feedback); agents[pid]["traj"].step_infos.append(step_info)
        if done: break
    final_rewards, game_info = env.close()
    for pid in agents.keys():
        if agents[pid]["traj"]!=None: agents[pid]["traj"].final_reward=final_rewards[pid]; agents[pid]["traj"].game_info=game_info[pid]; agents[pid]["traj"].num_turns=turn
        if game_info[pid]["invalid_move"] and agents[pid]["traj"]!=None: agents[pid]["traj"].format_feedbacks[-1]["invalid_move"]=True
    game_information.final_rewards=final_rewards; game_information.num_turns=turn; game_information.game_info=game_info
    return game_information, [agents[pid]["traj"] for pid in agents.keys() if agents[pid]["traj"]!=None]


@ray.remote
class Collector:
    def __init__(self, vllm_config, tracker, buffer, game_scheduler):
        self.logger = setup_logger("collector", ray.get(tracker.get_log_dir.remote()))
        self.tracker, self.buffer, self.game_scheduler = tracker, buffer, game_scheduler
        # self.actors = [VLLMActor.options(num_gpus=1).remote(cfg=vllm_config, tracker=tracker, name=f"Actor-{i}") for i in range(int(ray.available_resources().get("GPU", 0))-1)]
        self.actors = [VLLMActor.options(num_gpus=1).remote(cfg=vllm_config, tracker=tracker, name=f"Actor-{i}") for i in range(int(ray.available_resources().get("GPU", 0)))]
        self._actor_iter = itertools.cycle(self.actors)

        # thead keeping
        self.flight: Dict[ray.ObjectRef, TaskMeta] = {}
        self._num_running = lambda typ: sum(meta.type == typ for meta in self.flight.values())
        self.logger.info("Collector initialized")
    
    def _launch_jobs(self, max_train: int, max_eval: Optional[int]):
        while self._num_running("train") < max_train: # submit new train game
            try:
                game_spec: GameSpec = ray.get(self.game_scheduler.next_train_job.remote()) # sample game spec
                self.logger.info(f"received train game_spec: {game_spec}")
                actor: VLLMActor = next(self._actor_iter) # get actor
                ref = run_game.remote(game_spec, actor)
                self.flight[ref] = TaskMeta("train", game_spec.env_id)
            except Exception as exc:
                self.logger.info(f"Exception in train game {game_spec}: {exc}")

        while max_eval!=None and self._num_running("eval") < max_eval:
            try:
                game_spec: GameSpec = ray.get(self.game_scheduler.next_eval_job.remote())
                self.logger.info(f"received eval game_spec: {game_spec}")
                actor: VLLMActor = next(self._actor_iter) # get actor
                ref = run_game.remote(game_spec, actor)
                self.flight[ref] = TaskMeta("eval", game_spec.env_id)
            except Exception as exc:
                self.logger.info(f"Exception in eval game {game_spec}: {exc}")

    def _handle_finished_job(self, ref):
        meta = self.flight.pop(ref)
        try: game_information, player_trajs = ray.get(ref)
        except (RayTaskError, RayActorError) as err: self.logger.error(f"Remote episode failed for {meta.type} task: env={meta.env_id}: {err}", exc_info=True); return
        self._post_train(meta, game_information, player_trajs) if meta.type=="train" else self._post_eval(meta, game_information)
    
    def _post_train(self, meta: TaskMeta, game_information: GameInformation, player_trajs: List[PlayerTrajectory]):
        for traj in player_trajs: self.buffer.add_player_trajectory.remote(traj, env_id=meta.env_id); self.tracker.add_player_trajectory.remote(traj, env_id=meta.env_id)
        self.game_scheduler.update.remote(game_info=game_information)

    def _post_eval(self, meta: TaskMeta, game_information: GameInformation):
        self.tracker.add_eval_game_information.remote(game_information=game_information, env_id=meta.env_id)
    
    def collect(self, num_train_workers: int, num_eval_workers: Optional[int]=None):
        self.logger.info("entered collect func")
        while ray.get(self.buffer.continue_collection.remote()):
            self.logger.info("entered colelct loop")
            self._launch_jobs(num_train_workers, num_eval_workers)
            if not self.flight: continue
            done_ref, _ = ray.wait(list(self.flight), num_returns=1)
            self._handle_finished_job(done_ref[0])



import os, time, asyncio
from collections import defaultdict, deque
from typing import Optional, Dict, Any

import ray
from vllm import EngineArgs, LLMEngine, SamplingParams
from vllm.lora.request import LoRARequest

from unstable.utils.logging import setup_logger


@ray.remote
class VLLMActor:
    def __init__(self, cfg: Dict[str, Any], tracker, name: str):
        self.logger = setup_logger(f"actor-{name}", ray.get(tracker.get_log_dir.remote())) # set up logging
        self.gpu_ids = ray.get_gpu_ids()
        os.environ["CUDA_VISIBLE_DEVICES"] = ",".join(map(str, self.gpu_ids))
        
        engine_args = EngineArgs(
            model=cfg["model_name"], enable_lora=True, max_loras=cfg["max_loras"], max_lora_rank=cfg["lora_config"]["lora_rank"], 
            max_cpu_loras=cfg["max_loras"], max_num_seqs=cfg["max_parallel_seq"], task="generate", max_model_len=cfg["max_model_len"],
            disable_custom_all_reduce=True, enforce_eager=False, disable_log_stats=True,  # Reduce logging overhead
        )
        try: self.engine = LLMEngine.from_engine_args(engine_args); self.logger.info("VLLM engine initialized successfully")
        except Exception as e: self.logger.error(f"VLLM engine initialization failed: {e}"); raise
        self.logger.info(f"vLLM model path or name: {engine_args.model}")
        self.logger.info(f"Model architecture: {self.engine.model_config.__dict__}")
            
        self.sampling_params = SamplingParams(temperature=cfg.get("temperature", 0.7), top_p=cfg.get("top_p", 0.95), max_tokens=cfg.get("max_tokens", 4096))

        self._queue = deque()
        self._futures = {}
        self._next_id = 0
        self._req2lora = {}
        self._prev_tok_cnt = defaultdict(int)

        self.tracker = tracker
        self.name = name

        self._queued = 0
        self._running = 0
        self._tok_hist = deque()
        self._batch_task = asyncio.create_task(self._batch_loop())
        self._report_task = asyncio.create_task(self._report_loop())
        self._lora_ids: Dict[str, int] = {"base": 0}
        self._next_lora_id = 1
        self._last_step_time = time.monotonic()  # Add health check flag

    async def submit_prompt(self, prompt: str, lora_path: Optional[str] = None) -> str:
        if lora_path is not None and not isinstance(lora_path, str): lora_path = str(lora_path)
        fut = asyncio.Future()
        self._queued += 1
        self._queue.append((prompt, lora_path, fut))
        return await fut

    async def _batch_loop(self):
        while True:
            try:
                await asyncio.sleep(0.02)
                if time.monotonic() - self._last_step_time > 30: 
                    self.logger.error(f"Potential deadlock detected - no engine steps for {time.monotonic() - self._last_step_time:.1f} seconds\nRunning requests: {dict(self._running)}\nQueue size: {len(self._queue)}") # 30 second deadlock detection
                while self._queue:
                    prompt, path, fut = self._queue.popleft()
                    lora = path or "base"
                    req_id = str(self._next_id); self._next_id += 1
                    self._futures[req_id] = fut
                    self._req2lora[req_id] = lora
                    self._queued -= 1
                    self._running += 1
                    self.logger.info(f"path: {path}, lora_ids: {self._lora_ids}")

                    if path:
                        if path not in self._lora_ids:
                            self._lora_ids[path] = self._next_lora_id
                            self._next_lora_id += 1
                        lora_req = LoRARequest(path, self._lora_ids[path], path)
                        self.logger.info(lora_req)
                    else:
                        lora_req = None
                    self.logger.info(f"Using {lora_req}")
                    try:
                        self.engine.add_request(req_id, prompt, self.sampling_params, lora_request=lora_req)
                        self.logger.debug(f"Added request {req_id} with lora {lora}")
                    except Exception as e:
                        self.logger.error(f"Failed to add request {req_id}: {e}")
                        self._running -= 1
                        self._req2lora.pop(req_id, None)
                        fut.set_exception(e)
                        continue
                try:
                    step_start = time.monotonic()
                    outs = self.engine.step()
                    step_duration = time.monotonic() - step_start
                    self._last_step_time = time.monotonic()
                    if step_duration > 5.0: self.logger.warning(f"Slow engine step: {step_duration:.1f}s") # Log slow steps
                except Exception as exc:   
                    self.logger.exception(f"engine.step() failed - running: {dict(self._running)}"); await asyncio.sleep(1.0)  # Brief pause before retry
                    continue

                for out in outs:
                    req_id = out.request_id
                    lora = self._req2lora.get(req_id, "base")
                    segment = out.outputs[-1]

                    tok_ids = getattr(segment, "token_ids", None) or []
                    prev = self._prev_tok_cnt[req_id]
                    new_tok = max(0, len(tok_ids) - prev)
                    self._prev_tok_cnt[req_id] = len(tok_ids)

                    now = time.monotonic()
                    for _ in range(new_tok): 
                        self._tok_hist.append(now)
                    if segment.finish_reason is not None:
                        fut = self._futures.pop(req_id, None)
                        if fut and not fut.done():
                            fut.set_result(segment.text)
                        self._running -= 1
                        self._req2lora.pop(req_id, None)
                        self._prev_tok_cnt.pop(req_id, None)
            except Exception as e: self.logger.exception(f"Critical error in batch loop: {e}"); await asyncio.sleep(1.0)  # Prevent tight error loop

    async def _report_loop(self):
        self.logger.info("Starting _report_loop")
        while True:
            await asyncio.sleep(5.0) # only send every 5 sec
            stats = {"queued": self._queued, "running": self._running, "tok_s": self._tok_rate()}
            self.logger.info(f"inside while loop _report_loop stats: {stats}")
            try: ray.get(self.tracker.log_inference.remote(actor=self.name, gpu_ids=self.gpu_ids, stats=stats))
            except Exception as e: self.logger.warning(f"tracker logging failed: {e}")

    def _tok_rate(self, window: float = 2.0) -> float:
        now  = time.monotonic()
        while self._tok_hist and now - self._tok_hist[0] > window:
            self._tok_hist.popleft()
        return len(self._tok_hist) / window


from unstable.collector import Collector 
from unstable.buffers import StepBuffer, EpisodeBuffer
from unstable.trackers import Tracker
from unstable.learners import REINFORCELearner, A2CLearner
from unstable.terminal_interface import TerminalInterface
from unstable.model_registry import ModelRegistry
from unstable.game_scheduler import GameScheduler
from unstable._types import TrainEnvSpec, EvalEnvSpec
import unstable.samplers
import unstable.samplers.env_samplers
import unstable.samplers.model_samplers
from unstable.runtime import build

__all__ = ["build", "Collector", "StepBuffer", "EpisodeBuffer", "REINFORCELearner", "A2CLearner", "Tracker", "ModelRegistry", "GameScheduler", "TerminalInterface", "TrainEnvSpec", "EvalEnvSpec"]
__version__ = "0.2.0"


from typing import List
from unstable._types import PlayerTrajectory

class StepRewardTransform:
    def __call__(self, player_traj: PlayerTrajectory, step_index: int, reward: float) -> float: raise NotImplementedError

class ComposeStepRewardTransforms:
    def __init__(self, transforms: List[StepRewardTransform]): self.transforms = transforms
    def __call__(self, player_traj: PlayerTrajectory, step_index: int, reward: float) -> float:
        for transform in self.transforms: reward = transform(player_traj, step_index, reward)
        return reward

class RewardForFormat(StepRewardTransform):
    def __init__(self, reward: float=0, penalty: float=0): self.reward, self.penalty = reward, penalty
    def __call__(self, player_traj: PlayerTrajectory, step_index: int, reward: float) -> float:
        reward += (self.reward if player_traj.format_feedbacks[step_index].get("correct_answer_format") else self.penalty)
        return reward

class PenaltyForInvalidMove(StepRewardTransform):
    def __init__(self, reward: float=0, penalty: float=0): self.reward, self.penalty = reward, penalty
    def __call__(self, player_traj: PlayerTrajectory, step_index: int, reward: float) -> float:
        reward += (self.penalty if player_traj.format_feedbacks[step_index].get("invalid_move") else self.reward)
        return reward


import trueskill
from dataclasses import dataclass, field
from typing import List, Any, Dict, Optional


@dataclass
class Step:
    pid: int
    obs: str 
    act: str
    reward: float
    env_id: str
    step_info: Optional[Dict]

@dataclass
class PlayerTrajectory:
    pid:                int = field(default_factory=int)
    final_reward:       float = field(default_factory=float)
    obs:                List[str] = field(default_factory=list)
    actions:            List[str] = field(default_factory=list)
    extracted_actions:  List[str] = field(default_factory=list)
    format_feedbacks:   List[Dict] = field(default_factory=list)
    step_infos:         List[Dict] = field(default_factory=list)
    game_info:          Dict = field(default_factory=dict)
    num_turns:          int = field(default_factory=int)


@dataclass
class GameInformation:
    game_idx:           int = field(default_factory=int)
    pid:                List[int] = field(default_factory=list)
    obs:                List[str] = field(default_factory=list)
    full_actions:       List[str] = field(default_factory=list)
    extracted_actions:  List[str] = field(default_factory=list)
    step_infos:         List[Dict] = field(default_factory=list)
    game_info:          Dict = field(default_factory=dict)
    final_rewards:      Dict[int, float] = field(default_factory=dict)
    num_turns:          int = field(default_factory=int)
    names:              Dict[int, str] = field(default_factory=dict)
    eval_model_pid:     Optional[int] = None
    eval_opponent_name: Optional[str] = None

@dataclass 
class AgentSpec:
    pid: int
    kind: str # "checkpoint" | "openrouter"
    collect_data: bool = False
    openrouter_name: str|None = None
    lora_path: str|None = None
    prompt_template: str = "default" # prompt template key
    action_extraction_fn: str = "default"

@dataclass
class GameSpec:
    game_idx: int
    env_id: str
    seed: int
    agent_specs: List[AgentSpec]
    eval_model_pid: Optional[int] = None
    eval_opponent_name: Optional[str] = None


@dataclass
class TaskMeta:
    type: str  # "train" | "eval"
    env_id: str


@dataclass
class TrainEnvSpec:
    env_id: str
    num_players: int
    num_actors: int
    prompt_template: str 
    action_extraction_fn: str = "default"

@dataclass 
class EvalEnvSpec:
    env_id: str 
    num_players: int 
    prompt_template: str
    action_extraction_fn: str = "default"
    fixed_opponent: str = "google/gemini-2.0-flash-lite-001"
    # forced_pid: Optional[List] = None # whether to force a specific pid for the collection models


@dataclass
class ModelMeta:
    uid: str
    kind: str # "checkpoint" | "fixed"
    path_or_name: str # local path or OpenRouter id
    rating: trueskill.Rating # μ / σ
    games: int = 0
    wins: int = 0
    draws: int = 0
    active: bool = True
    iteration: int|None = None





import ray, torch, tree, random
from typing import Optional
from dataclasses import replace
from unstable.learners.base import BaseLearner
from unstable.learners.utils import build_peft_model, enable_full_activation_ckpt
from unstable.reward_transformations.transformation_sampling import NormalizeRewardsByEnv

def compute_gae(rewards, values, gamma=1.0, gae_lambda=1.0): # Compute gae (for policy learning) and return (for critic learning)
    assert len(rewards) == len(values)
    advantages = torch.zeros_like(rewards)
    lastgaelam = 0
    for t in reversed(range(len(rewards))):
        if t == len(rewards) - 1:
            nextnonterminal = 0  # Assume a complete episode
            nextvalues = 0  # Does not matter
        else:
            nextnonterminal = 1.0
            nextvalues = values[t + 1]
        delta = rewards[t] + gamma * nextvalues * nextnonterminal - values[t]
        advantages[t] = lastgaelam = delta + gamma * gae_lambda * nextnonterminal * lastgaelam
    return advantages

@ray.remote
class A2CLearner(BaseLearner):
    def initialize_algorithm(self, infer_mini_batch_size: int, critic_learning_rate: float, normalize_adv: bool=False, max_generation_len: Optional[int]=None, max_train_len: Optional[int]=None, initial_lora_path: Optional[str]=None):
        self.infer_mini_batch_size = infer_mini_batch_size
        self.normalize_adv = normalize_adv
        self.max_train_len = max_train_len
        self.max_generation_len = max_generation_len

        # build the critic
        self.critic, _ = build_peft_model(self.model_name, self.device, self.lora_cfg, initial_lora_path, critic_model=True)
        if not self.use_trainer_cache:      self.policy_model.config.use_cache = False
        if self.gradient_checkpointing:     self.policy_model.gradient_checkpointing_enable() # gradient checkpointing
        if self.activation_checkpointing:   enable_full_activation_ckpt(self.policy_model)       # activation checkpointing. Affords most of the vRAM savings
        self.critic_optimizer = torch.optim.AdamW(filter(lambda p: p.requires_grad, self.critic.parameters()), lr=critic_learning_rate,)

    def _prepare_batch(self, steps):
        obs, acts, advs, rets = zip(*[(s.obs, s.act, s.reward, s.step_info.get("return", torch.nan)) for s in steps])
        advs = torch.tensor(advs, dtype=torch.float32, device=self.device)
        rets = torch.tensor(rets, dtype=torch.float32, device=self.device)
        combined = [o + a for o, a in zip(obs, acts)]
        lengths = [len(self.tokenizer(text, add_special_tokens=False)["input_ids"]) for text in combined]
        avg_len = sum(lengths) / len(lengths)
        pct_truncated = (sum(l > self.max_train_len for l in lengths) / len(lengths) if self.max_train_len else 0)
        enc = self.tokenizer(combined, return_tensors="pt", padding=True, truncation=True, max_length=self.max_train_len).to(self.device)  # Tokenize with truncation
        state_enc = self.tokenizer(obs, return_tensors="pt", padding=True, truncation=True, max_length=self.max_train_len).to(self.device)  # Tokenize with truncation
        return enc, state_enc, advs, rets, obs, avg_len, pct_truncated

    def _mini_batch_update_step(self, steps, scaling: float = 1.0):
        enc, state_enc, advs, rets, obs, avg_len, pct_truncated = self._prepare_batch(steps=steps)
        # Learn policy
        out = self.policy_model(**enc)
        logp = torch.nn.functional.log_softmax(out.logits, dim=-1)
        tgt_ids = enc.input_ids[:, 1:]
        tok_logp = logp[:, :-1, :].gather(-1, tgt_ids.unsqueeze(-1)).squeeze(-1)
        mask = torch.ones_like(enc.input_ids, dtype=torch.bool, device=self.device)  # build prompt mask
        for i, o in enumerate(obs): mask[i, : len(self.tokenizer(o, add_special_tokens=False)["input_ids"])] = False
        mask = mask[:, 1:]
        seq_logp = (tok_logp * mask).sum(1) / self.max_generation_len
        loss = -(advs * seq_logp).mean() / scaling
        loss.backward()
        torch.cuda.empty_cache()

        # Learn value
        value_pred = self.critic(**state_enc)[:, 0]
        value_loss = 0.5 * ((value_pred - rets) ** 2).mean()
        value_loss.backward()
        torch.cuda.empty_cache()

        return {
            "policy_loss": loss.item(), "value_loss": value_loss.item(), "logp_mean": seq_logp.mean().item(), "value_mae": (value_pred-rets).abs().mean().item(), "value_dir_acc": ((value_pred > 0) == (rets > 0)).float().mean().item(),
            "logp_std": seq_logp.std().item(), "num_steps": len(steps), "avg_train_len": avg_len, "pct_truncated": pct_truncated,
        }


    def _update(self, batch):
        all_samples = tree.flatten(batch)
        num_samples = len(all_samples)

        # construct value targets
        all_values = []
        for i in range(0, len(all_samples), self.infer_mini_batch_size):
            batch_steps = all_samples[i : i + self.infer_mini_batch_size]
            with torch.autocast(device_type=self.device.type, dtype=torch.bfloat16):
                _, state_enc, _, _, _, _, _ = self._prepare_batch(batch_steps)
                with torch.no_grad():
                    # take the first token's output
                    target_values = self.critic(**state_enc)[:, 0]
                    all_values.append(target_values)
        all_values = torch.cat(all_values).float().cpu()
        ep_values = torch.split(all_values, [len(ep) for ep in batch])
        ep_rewards = [torch.tensor([step.reward for step in ep]) for ep in batch]
        ep_advantages = []
        ep_returns = []
        for rewards, values in zip(ep_rewards, ep_values):
            adv = compute_gae(rewards, values)
            ep_advantages.append(adv)
            ep_returns.append(adv + values)

        train_batch = []
        for i, ep in enumerate(batch):
            for j, step in enumerate(ep):
                step = replace(step, reward=ep_advantages[i][j].item())
                step = replace(step, step_info={**step.step_info, "return": ep_returns[i][j].item(), "advantage": ep_advantages[i][j].item()})
                train_batch.append(step)
        assert len(train_batch) >= self.batch_size

        self.policy_optimizer.zero_grad(set_to_none=True)
        self.critic_optimizer.zero_grad(set_to_none=True)

        metrics_acc: Dict[str, float] = {}

        train_batch = random.sample(train_batch, self.batch_size)
        num_steps = self.batch_size // self.mini_batch_size
        self.logger.info(f"Got {num_samples} many samples. Running for {num_steps} steps (i.e. mini batch size: {self.mini_batch_size})")

        if self.normalize_adv: train_batch = NormalizeRewardsByEnv(True)(train_batch)
        for i in range(num_steps):
            sub = train_batch[i * self.mini_batch_size : (i + 1) * self.mini_batch_size]
            with torch.autocast(device_type=self.device.type, dtype=torch.bfloat16):
                update_metrics = self._mini_batch_update_step(sub, scaling=num_steps)
            for k, v in update_metrics.items(): metrics_acc[k] = metrics_acc.get(k, 0.0) + v
            self.logger.info(f"Mini-step metrics: {update_metrics}")
        self.logger.info(f"Step metrics: {metrics_acc}")

        torch.nn.utils.clip_grad_norm_(self.policy_model.parameters(), self.grad_clip)
        torch.nn.utils.clip_grad_norm_(self.critic.parameters(), self.grad_clip)

        try:
            self.policy_optimizer.step()
            self.critic_optimizer.step()
        except Exception as exc:
            self.logger.exception(f"optimizer.step crashed on step {self._step} -\n\n{exc}\n\n")
            raise
        self._step += 1

        log = {f"{k}": v / num_steps for k, v in metrics_acc.items()}
        grad_norm = (sum(p.grad.data.norm(2).item() ** 2 for p in self.policy_model.parameters() if p.grad is not None) ** 0.5)
        critic_grad_norm = (sum(p.grad.data.norm(2).item() ** 2 for p in self.critic.parameters() if p.grad is not None)** 0.5)
        log.update({
            "step": self._step, "samples_seen": self._samples_seen, "lr": self.policy_optimizer.param_groups[0]["lr"], 
            "grad_norm": grad_norm, "critic_lr": self.critic_optimizer.param_groups[0]["lr"], "critic_grad_norm": critic_grad_norm,
        })
        return log










import ray, random
from unstable.utils import setup_logger
from unstable._types import AgentSpec, GameSpec, GameInformation

@ray.remote
class GameScheduler:
    def __init__(self, model_sampler, env_sampler, logging_dir: str):
        self.logger = setup_logger("game_scheduler", logging_dir)
        self.model_sampler = model_sampler
        self.env_sampler = env_sampler
        self._game_idx = 0
        self._running_jobs = {}

    def next_train_job(self):
        try:
            env_spec = self.env_sampler.sample(kind="train") # sample the env spec
            current_ckpt_uid, current_ckpt_lora_path = self.model_sampler.get_current_ckpt() # sample the current checkpoint
            # build the game spec and agent specs
            pids = list(range(env_spec.num_players))
            random.shuffle(pids); agent_specs = []
            self._running_jobs[self._game_idx] = {"env_id": env_spec.env_id, "models": []}
            for i, pid in enumerate(pids):
                if i < env_spec.num_actors: # add current ckpt
                    self._running_jobs[self._game_idx]["models"].append({"uid": current_ckpt_uid, "pid": pid, "type": "model"})
                    agent_specs.append(AgentSpec(pid=pid, kind="checkpoint", collect_data=True, lora_path=current_ckpt_lora_path, prompt_template=env_spec.prompt_template, action_extraction_fn=env_spec.action_extraction_fn))
                else: # sample opponent and add
                    opp_uid, kind, opp_lora_path, opp_openrouter_name = self.model_sampler.sample_opponent()
                    agent_specs.append(AgentSpec(pid=pid, kind=kind, lora_path=opp_lora_path, openrouter_name=opp_openrouter_name)) # TODO might have to adjust what is passed
                    self._running_jobs[self._game_idx]["models"].append({"uid": opp_uid, "pid": pid, "type": "opponent"})
            game_spec = GameSpec(game_idx=self._game_idx, env_id=env_spec.env_id, seed=self._game_idx, agent_specs=agent_specs) # populate GameSpec
            self._game_idx += 1
            return game_spec
        except Exception as exc:
            self.logger.info(f"Exception in 'next_train_job': {exc}")
            import time 
            time.sleep(500)

    def next_eval_job(self):
        try:
            env_spec = self.env_sampler.sample(kind="eval")
            current_ckpt_uid, current_ckpt_lora_path = self.model_sampler.get_current_ckpt() # sample the current checkpoint
            pids = list(range(env_spec.num_players))
            random.shuffle(pids); agent_specs = []
            for i, pid in enumerate(pids):
                if i == 0:  agent_specs.append(AgentSpec(pid=pid, kind="checkpoint", collect_data=True, lora_path=current_ckpt_lora_path, prompt_template=env_spec.prompt_template, action_extraction_fn=env_spec.action_extraction_fn)) # only one actor, rest fixed
                else:       agent_specs.append(AgentSpec(pid=pid, kind="openrouter", lora_path=None, openrouter_name=env_spec.fixed_opponent)) # sample opponent and add
            game_spec = GameSpec(game_idx=self._game_idx, env_id=env_spec.env_id, seed=self._game_idx, agent_specs=agent_specs, eval_model_pid=pids[0], eval_opponent_name=env_spec.fixed_opponent) # populate GameSpec
            return game_spec
        except Exception as exc:
            self.logger.info(f"Exception in 'next_eval_job': {exc}")
            import time 
            time.sleep(500)

    def update(self, game_info: GameInformation):
        job_info = self._running_jobs.pop(game_info.game_idx, None)
        if job_info is None: return # shouldn’t happen
        actor_rs = [game_info.final_rewards[m["pid"]] for m in job_info["models"] if m["type"] == "model" if m["pid"] in game_info.final_rewards]
        opp_rs = [game_info.final_rewards[m["pid"]] for m in job_info["models"] if m["type"] == "opponent" if m["pid"] in game_info.final_rewards]
        self.env_sampler.update(avg_actor_reward=(sum(actor_rs) / len(actor_rs) if actor_rs else None), avg_opponent_reward=(sum(opp_rs) / len(opp_rs) if opp_rs else None)) # update environment sampling 
        self.model_sampler.update(game_info=game_info, job_info=job_info) # update model sampler




import re
from typing import Tuple, Dict, Callable


def format_template(system: str = "", user: str = "", assistant: str = "") -> str: return f"{system}{user}{assistant}"
TEMPLATE_PARTS = {
    "default": {
        "user": lambda obs: f"You are playing a two-player zero-sum game. Make valid moves to win. You should first reason about your next move, and then submit the move enclosed by \\boxed{{}}.\nObservation: {obs}\n"
    },
    "qwen3-zs": {
        "user": lambda obs: f"<|im_start|>user\nYou are playing a two-player zero-sum game. Make valid actions to win.\nObservation: {obs}\nPlease reason step by step, and put your final answer within \\boxed{{}}.<|im_end|>\n",
        "assistant": "<|im_start|>assistant\n"
    },
    "qwen3-sp": {
        "user": lambda obs:  f"<|im_start|>user\nYou are playing a single-player game. Make valid actions to solve it completely.\nObservation: {obs}\nPlease reason step by step, and put your final answer within \\boxed{{}}.<|im_end|>\n",
        "assistant": "<|im_start|>assistant\n"
    },
    "qwen3-reasoning": {
        "user": lambda obs: f"<|im_start|>user\nPlease reason step by step, and put your final answer within \\boxed{{}}.\nQuestion: {obs}<|im_end|>\n",
        "assistant": "<|im_start|>assistant\n<think>"
    },
    "gemma3-zs": {
        "user": lambda obs: f"<bos><start_of_turn>user\nYou are playing a two-player zero-sum game. Make valid actions to win.\nObservation: {obs}\nPlease reason step by step, and put your final answer within \\boxed{{}}.<end_of_turn>\n",
        "assistant": "<start_of_turn>model\n"
    },
    "llama-instruct-zs": {
        "system": "<|begin_of_text|><|start_header_id|>system<|end_header_id|>\n\nYou are playing a two-player zero-sum game. Make valid actions to win.<|eot_id|>",
        "user": lambda obs: f"<|start_header_id|>user<|end_header_id|>\n\nCurrent Observation: {obs}\nPlease reason step by step, and put your final answer within \\boxed{{}}.<|eot_id|>\n",
        "assistant": "<|start_header_id|>assistant<|end_header_id|>"
    },
}
def apply_template(template_name: str, observation: str) -> str:
    parts = TEMPLATE_PARTS.get(template_name)
    return format_template(system=parts.get("system", ""), user=parts["user"](observation), assistant=parts.get("assistant", ""))


def extract_action_and_format_feedback(raw_action: str) -> Tuple[str, Dict[str, bool]]:
    matches = re.findall(r"\\boxed\{(.*?)\}", raw_action)
    if matches:
        last_match = matches[-1].strip()
        if last_match:  # non-empty boxed
            action = f"[{last_match}]" if "[" not in last_match else last_match
            has_think = 1
        else:  # empty boxed
            action = raw_action
            has_think = 0
    else:  # no boxed at all
        action = raw_action
        has_think = 0

    format_feedback = {"correct_answer_format": bool(has_think)}
    return action, format_feedback


OBSERVATION_FORMATTING: Dict[str, Callable[[str], str]] = {key: (lambda key=key: lambda observation: apply_template(key, observation))() for key in TEMPLATE_PARTS}
ACTION_EXTRACTION = {"default": extract_action_and_format_feedback}


import copy
from typing import Dict, List, Tuple, Union, Optional

class FinalRewardTransform:
    def __call__(self, reward: float, pid: int, env_id: Optional[str] = None) -> float: raise NotImplementedError

class ComposeFinalRewardTransforms:
    def __init__(self, transforms: List[FinalRewardTransform]): self.transforms = transforms
    def __call__(self, reward: float, pid: int, env_id: Optional[str] = None) -> float:
        for transform in self.transforms: reward = transform(reward, pid, env_id)
        return reward

class RoleAdvantageFormatter(FinalRewardTransform):
    def __init__(self, role_adv: float=0.0, tau: float=0.001): self.role_adv, self.tau = role_adv, tau
    def __call__(self, reward: float, pid: int, env_id: Optional[str] = None) -> float:
        self.role_adv[pid] = (1-self.tau) * self.role_adv[pid] + self.tau * reward
        reward -= self.role_adv[pid]
        return reward

class RoleAdvantageByEnvFormatter(FinalRewardTransform):
    def __init__(self, default_role_adv: float=0.0, tau: float=0.001): self.default_role_adv, self.tau, self.role_advantage_dict = default_role_adv, tau, {}
    def __call__(self, reward: float, pid: int, env_id: Optional[str] = None) -> float:
        if env_id not in self.role_advantage_dict: self.role_advantage_dict[env_id] = {}
        self.role_advantage_dict[env_id][pid] = (1-self.tau) *  self.role_advantage_dict[env_id].get(pid, self.default_role_adv) + self.tau * reward
        reward -= self.role_advantage_dict[env_id][pid]
        return reward



import ray, torch
from unstable.learners.base import BaseLearner

@ray.remote
class REINFORCELearner(BaseLearner):
    def initialize_algorithm(self, max_train_len: int, max_generation_len: int):
        self.max_train_len = max_train_len
        self.max_generation_len = max_generation_len # for Dr. GRPO trick

    def _prepare_batch(self, steps):
        obs, acts, advs = zip(*[(s.obs, s.act, s.reward) for s in steps])
        advs = torch.tensor(advs, dtype=torch.float32, device=self.device)
        combined = [o + a for o, a in zip(obs, acts)]
        lengths = [len(self.tokenizer(text, add_special_tokens=False)["input_ids"]) for text in combined]
        avg_len = sum(lengths) / len(lengths)
        pct_truncated = sum(l > self.max_train_len for l in lengths) / len(lengths) if self.max_train_len else 0
        enc = self.tokenizer(combined, return_tensors="pt", padding=True, truncation=True, max_length=self.max_train_len).to(self.device) # Tokenize with truncation
        return enc, advs, obs, avg_len, pct_truncated

    def _mini_batch_update_step(self, steps, scaling: float = 1.0):
        enc, advs, obs, avg_len, pct_truncated = self._prepare_batch(steps=steps)
        out = self.policy_model(**enc)
        logp = torch.nn.functional.log_softmax(out.logits, dim=-1)
        tgt_ids = enc.input_ids[:, 1:]
        tok_logp = logp[:, :-1, :].gather(-1, tgt_ids.unsqueeze(-1)).squeeze(-1)
        mask = torch.ones_like(enc.input_ids, dtype=torch.bool, device=self.device) # build prompt mask
        for i, o in enumerate(obs): mask[i, :len(self.tokenizer(o, add_special_tokens=False)["input_ids"])] = False
        mask = mask[:, 1:]
        seq_logp = (tok_logp * mask).sum(1) / self.max_generation_len
        loss = -(advs * seq_logp).mean() / scaling
        loss.backward()
        torch.cuda.empty_cache()
        return {"loss": loss.item(), "logp_mean": seq_logp.mean().item(), "avg_train_len": avg_len, "pct_truncated": pct_truncated}
    
    def _update(self, batch):
        metrics_acc = {}
        self.policy_optimizer.zero_grad(set_to_none=True)
        for i in range(self.gradient_acc_steps):
            sub = batch[i * self.mini_batch_size : (i + 1) * self.mini_batch_size]
            with torch.autocast(device_type=self.device.type, dtype=torch.bfloat16): 
                update_metrics = self._mini_batch_update_step(sub, scaling=self.gradient_acc_steps)
            for k, v in update_metrics.items(): metrics_acc[k] = metrics_acc.get(k, 0.0) + v /  self.gradient_acc_steps
            self.logger.info(f"Mini-step metrics: {update_metrics}")
        self.logger.info(f"Step metrics: {metrics_acc}")
        torch.nn.utils.clip_grad_norm_(self.policy_model.parameters(), self.grad_clip)
        self.policy_optimizer.step()
        return metrics_acc
        


import random
from typing import Protocol, List, Any
from unstable._types import TrainEnvSpec, EvalEnvSpec, GameInformation


class BaseEnvSampler: 
    def __init__(self, train_env_specs: List[TrainEnvSpec], eval_env_specs: List[EvalEnvSpec]|None = None, rng_seed: int|None = 489):
        assert train_env_specs, "Need at least one training environment"
        self._train, self._eval = train_env_specs, eval_env_specs
        self._rng = random.Random(rng_seed)
    def env_list(self) -> str: return ",".join([tes.env_id for tes in self._train])
    def sample(self, kind: str = "train") -> TrainEnvSpec | EvalEnvSpec: ...
    def update(self, avg_actor_reward: float, avg_opponent_reward: float|None) -> None: ...

class UniformRandomEnvSampler(BaseEnvSampler):
    def sample(self, kind: str = "train") -> TrainEnvSpec | EvalEnvSpec:
        if kind not in ("train", "eval"): raise ValueError(f"kind must be 'train' or 'eval' (got {kind!r})")
        return self._rng.choice(self._train if kind == "train" else self._eval)
    def update(self, avg_actor_reward: float, avg_opponent_reward: float|None) -> None: return # No-op for the uniform strategy - but the hook is here so the scheduler can call it unconditionally.




import csv
from unstable._types import GameInformation


def write_training_data_to_file(batch, filename: str):
    with open(filename, mode='w', newline='', encoding='utf-8') as csvfile:
        writer = csv.writer(csvfile)
        writer.writerow(['pid', 'obs', 'act', 'reward', "env_id", "step_info"])  # header
        for step in batch: writer.writerow([step.pid, step.obs, step.act, step.reward, step.env_id, step.step_info])

def write_game_information_to_file(game_info: GameInformation, filename: str) -> None:
    with open(filename, "w", newline="", encoding="utf-8") as f:
        writer = csv.DictWriter(f, fieldnames=["game_idx", "turn_idx", "pid", "name", "obs", "full_action", "extracted_action", "step_info", "final_reward", "eval_model_pid", "eval_opponent_name"])
        writer.writeheader()
        for t in range(game_info.num_turns or len(game_info.obs)):
            pid = game_info.pid[t] if t < len(game_info.pid) else None
            row = {"game_idx": game_info.game_idx, "turn_idx": t, "pid": pid, "name": game_info.names.get(pid, ""), "obs": game_info.obs[t] if t < len(game_info.obs) else "", "full_action": game_info.full_actions[t] if t < len(game_info.full_actions) else "", "extracted_action": game_info.extracted_actions[t] if t < len(game_info.extracted_actions) else "", "step_info": json.dumps(game_info.step_infos[t] if t < len(game_info.step_infos) else {}, ensure_ascii=False), "final_reward": game_info.final_rewards.get(pid, ""), "eval_model_pid": game_info.eval_model_pid, "eval_opponent_name": game_info.eval_opponent_name}
            writer.writerow(row)



import logging, os, pathlib
from rich.logging import RichHandler


def setup_logger(name: str, log_dir: str, level: int = logging.INFO, to_console: bool = False) -> logging.Logger:
    path = pathlib.Path(log_dir) / f"{name}.log"; path.parent.mkdir(parents=True, exist_ok=True)
    logger = logging.getLogger(name); logger.setLevel(level)
    logger.propagate = False # keep noise out of the root logger
    fh = logging.handlers.RotatingFileHandler(path, maxBytes=10_000_000, backupCount=5, encoding="utf-8")
    fh.setFormatter(logging.Formatter("%(asctime)s [%(levelname)s] %(name)s: %(message)s"))
    logger.addHandler(fh)
    if to_console: # optional colourful console for the *driver* only
        ch = RichHandler(rich_tracebacks=True, markup=True)
        ch.setFormatter(logging.Formatter("%(message)s"))
        logger.addHandler(ch)
    return logger



import ray, random
from typing import Dict, Any, Tuple, Optional

from unstable._types import GameInformation


class BaseModelSampler:
    def __init__(self, model_registry): 
        self.model_registry = model_registry
    
    def get_current_ckpt(self):         
        current_ckpt_uid = ray.get(self.model_registry.get_current_ckpt.remote())
        current_ckpt_lora_path = ray.get(self.model_registry.get_name_or_lora_path.remote(uid=current_ckpt_uid))
        return current_ckpt_uid, current_ckpt_lora_path
    
    def update(self, game_info: GameInformation, job_info: Dict[str, Any]):
        self.model_registry.update_ratings.remote(
            uids = [m["uid"] for m in job_info["models"] if m["pid"] in game_info.final_rewards],
            scores = [game_info.final_rewards[m["pid"]] for m in job_info["models"] if m["pid"] in game_info.final_rewards],
            env_id = job_info["env_id"]
        )
    def sample_opponent(self): raise NotImplementedError 

class FixedOpponentModelSampler(BaseModelSampler):
    def sample_opponent(self): 
        opponent_meta = random.choice([model_meta for uid, model_meta in ray.get(self.model_registry.get_all_models.remote()).items() if model_meta.active and model_meta.kind=="fixed"])
        return opponent_meta.uid, opponent_meta.kind, None, opponent_meta.path_or_name




import numpy as np
from typing import List, Optional
from collections import defaultdict
from unstable._types import Step

class SamplingRewardTransform:
    def __call__(self, steps: List[Step], env_id: Optional[str] = None) -> List[Step]: raise NotImplementedError

class ComposeSamplingRewardTransforms:
    def __init__(self, transforms: List[SamplingRewardTransform]):  self.transforms = transforms
    def __call__(self, steps: List[Step]) -> List[Step]:
        for transform in self.transforms: steps = transform(steps)
        return steps

class NormalizeRewards(SamplingRewardTransform):
    def __init__(self, z_score: bool=False): self.z_score = z_score
    def __call__(self, steps: List[Step], env_id: Optional[str] = None) -> List[Step]:
        rewards = [step.reward for step in steps]
        mean, std = np.mean(rewards), np.std(rewards)+1e-8
        for step in steps: step.reward = (step.reward-mean)/(std if self.z_score else 1)
        return steps

class NormalizeRewardsByEnv(SamplingRewardTransform):
    def __init__(self, z_score: bool = False): self.z_score = z_score 
    def __call__(self, steps: List[Step], env_id: Optional[str] = None) -> List[Step]:
        env_buckets = defaultdict(list)
        for step in steps: env_buckets[step.env_id].append(step) # bucket by env
        for env_steps in env_buckets.values():
            r = np.asarray([s.reward for s in env_steps], dtype=np.float32)
            normed = ((r-r.mean())/r.std()+1e-8) if self.z_score else r-r.mean()
            for s, nr in zip(env_steps, normed): s.reward = float(nr) # write back
        return steps




import ray, copy, trueskill
from dataclasses import dataclass, asdict
from collections import defaultdict
from typing import Dict, List

from unstable._types import ModelMeta
from unstable.utils import setup_logger


@ray.remote
class ModelRegistry:
    def __init__(self, tracker, beta: float = 4.0):
        self.TS = trueskill.TrueSkill(beta=beta)
        self._db: dict[str, ModelMeta] = {}
        self._match_counts = defaultdict(int) # (uid_a, uid_b) -> n
        self._exploration = defaultdict(lambda: defaultdict(dict))
        self._current_ckpt_uid : str | None = None 
        self._tracker = tracker; self._update_step: int = 1
        self.logger = setup_logger("model_registry", ray.get(self._tracker.get_log_dir.remote()))

    @staticmethod
    def _scores_to_ranks(scores: List[float]) -> List[int]:
        order = sorted(range(len(scores)), key=scores.__getitem__, reverse=True)
        ranks = [0]*len(scores); rank = 0
        for i, idx in enumerate(order):
            if i and scores[idx] != scores[order[i-1]]: rank = i  # next rank starts here
            ranks[idx] = rank
        return ranks

    def add_checkpoint(self, uid: str, path: str, iteration: int, inherit: bool=True):
        self.logger.info(f"tryin to add ckpt: {uid}, path {path}, iteration {iteration}, inherit: {inherit}")
        if uid in self._db: return
        rating = self.TS.Rating(mu=self._db[self._current_ckpt_uid].rating.mu, sigma=self._db[self._current_ckpt_uid].rating.sigma*2) if (inherit and self._current_ckpt_uid in self._db) else self.TS.create_rating()
        self._db[uid] = ModelMeta(uid=uid, kind="checkpoint", path_or_name=path, rating=rating, iteration=iteration)
        self._current_ckpt_uid = uid # make it current
        self.logger.info(f"added ckpt: {uid}, path {path}, iteration {iteration}, inherit: {inherit}")

    def get_all_models(self): return copy.deepcopy(self._db)
    def get_current_ckpt(self) -> str|None: return self._current_ckpt_uid
    def get_name_or_lora_path(self, uid: str) -> str: return self._db[uid].path_or_name
    def add_fixed(self, name: str, prior_mu: float = 25.): 
        if f"fixed-{name}" not in self._db: self._db[f"fixed-{name}"] = ModelMeta(f"fixed-{name}", "fixed", name, self.TS.create_rating(mu=prior_mu))

    def update_ratings(self, uids: List[str], scores: List[float], env_id: str, dummy_uid: str="fixed-env") -> None:
        if len(uids) == 1:
            if dummy_uid not in self._db: self.add_fixed(name=dummy_uid.replace("fixed-", ""), prior_mu=25.0)
            uids = [uids[0], dummy_uid]
            scores = [scores[0], 0.0] # any baseline score works
        rating_groups = [[self._db[uid].rating] for uid in uids]
        ranks = self._scores_to_ranks(scores)
        new_groups = self.TS.rate(rating_groups, ranks=ranks)

        # flatten, then write back
        for uid, (new_rating,) in zip(uids, new_groups):
            self._db[uid].rating = new_rating
            self._db[uid].games += 1
            if ranks[uids.index(uid)] == 0:               self._db[uid].wins  += 1
            elif ranks.count(ranks[uids.index(uid)]) > 1: self._db[uid].draws += 1

        # update pair-wise match matrix for analysis/debugging
        for i, uid_i in enumerate(uids):
            for uid_j in uids[i+1:]:
                self._match_counts[tuple(sorted((uid_i, uid_j)))] += 1
        self._update_step += 1

        # push to tracker every n update steps
        if not self._update_step%10: self._tracker.log_model_registry.remote(ts_dict={uid: asdict(meta) for uid, meta in self._db.items()}, match_counts=copy.deepcopy(self._match_counts))




import ray, asyncio, psutil, time, re, pynvml
from typing import Dict, Any, Tuple
import collections
from collections import deque
from itertools import zip_longest

from rich import box
from rich.text import Text
from rich.live import Live
from rich.table import Table
from rich.style import Style
from rich.color import Color
from rich.panel import Panel
from rich.layout import Layout
from rich.console import Console, Group

# TODO track the learner info here as well maybe

_HIST_BARS = " ▁▂▃▄▅▆▇█"
_UID_FIELD = 22
PALETTE = {"ok": "#A9B665", "warn": "#E0C06F", "crit": "#EA6962"}

def _bar(pct: float, width: int) -> str: return "█" * int(pct / 100 * width)
def _trim_uid(uid: str, width: int = _UID_FIELD) -> str: return uid if len(uid) <= width else f"…{uid[-width-2:]}"#f"{uid[:width//2-1]}…{uid[-(width//2-1)-3:]}"
def _scaled_hist(hist: deque[float|int], width: int) -> str:
    if not hist: return " " * width
    step = max(1, len(hist) // width); buckets = [list(hist)[i:i+step] for i in range(0, len(hist), step)]; result = ""
    for bucket in buckets[:width]:
        if not bucket: result += " "; continue
        level = int(((sum(bucket) / len(bucket)) - min(hist)) / (max(hist) - min(hist) if max(hist) != min(hist) else 1e-5) * (len(_HIST_BARS) - 1))
        result += _HIST_BARS[level]
    return result.ljust(width)

class TerminalInterface:
    def __init__(self, tracker, buffer):
        self.tracker, self.buffer = tracker, buffer; self.console = Console(color_system="truecolor") 
        self._gpu_stats, self._general_stats, self._tracker_stats = None, None, None
        self._hist: dict[str, deque[float|int]] = collections.defaultdict(lambda: deque(maxlen=128))
        self._max_tok_s: int = 1_000; pynvml.nvmlInit(); self.pynvml=pynvml; self.gpu_count=pynvml.nvmlDeviceGetCount()

        # Initialize panel attributes with placeholder panels
        self._gpu_panel = Panel(Text("waiting …"), title="GPU", box=box.SQUARE); self._base_stats_panel = Panel(Text("waiting …"), title="Collection Stats.", box=box.SQUARE)
        self._ts_panel = Panel(Text("waiting …"), title="TrueSkill", box=box.DOUBLE); self._heatmap_panel = Panel(Text("waiting …"), title="Match Frequencies", box=box.SQUARE)
        self._exploration_panel = Panel(Text("Not Implemented"), title="Exploration", box=box.DOUBLE)

    async def _system_stats(self) -> Dict[str, Any]:
        gpus = []
        if self.pynvml:
            for gid in range(self.gpu_count):
                h=self.pynvml.nvmlDeviceGetHandleByIndex(gid); power=self.pynvml.nvmlDeviceGetPowerUsage(h)/1000.0; limit=self.pynvml.nvmlDeviceGetEnforcedPowerLimit(h)/1000.0; m=self.pynvml.nvmlDeviceGetMemoryInfo(h)
                gpus.append({"id": gid, "used": m.used/1e9, "total": m.total/1e9, "mem_pct": (m.used/1e9)/(m.total/1e9)*100, "power": power, "limit": limit, "power_pct": power/limit*100 if limit else 0.0})
        return gpus

    async def _fetch_loop(self, interval: float = 2.0):
        while True:
            try:
                self._gpu_stats=await self._system_stats(); self._buffer_size=await self.buffer.size.remote(); self._tracker_stats=await self.tracker.get_interface_info.remote() # Fetch all stats
                self._gpu_panel = self._gpu(); self._base_stats_panel = self._base_stats(); self._ts_panel = self._ts(); self._heatmap_panel = self._heatmap(); self._exploration_panel = self._exploration() # Update all panels with new data
            except Exception as e: self.console.log(f"[red]stat-fetch error: {e}")
            await asyncio.sleep(interval)

    def _colour_for_util(self, pct: float) -> str: return PALETTE["ok"] if pct >= 75 else PALETTE["warn"] if pct >= 40 else PALETTE["crit"]
    def _gpu(self) -> Panel:
        if not self._gpu_stats or not self._tracker_stats: return Panel(Text("waiting …"), title="GPU", box=box.SQUARE)
        gpu_panels = []; bar_w = max(10, int(self.console.size.width * 0.48 - 20))
        for gpu_d in self._gpu_stats:
            tok_s = self._tracker_stats["gpu_tok_s"].get(gpu_d["id"], 0); self._max_tok_s = self._max_tok_s if tok_s<self._max_tok_s else tok_s
            tok_pct = tok_s/self._max_tok_s * 100; role = "Actor" if tok_s and tok_s > 0 else "Learner"
            line1 = Text.assemble(("TOK ", "dim"), Text(_bar(tok_pct, bar_w),            style=self._colour_for_util(tok_pct)),             f" {tok_s:5.0f} tok/s")
            line2 = Text.assemble(("PWR ", "dim"), Text(_bar(gpu_d['power_pct'], bar_w), style=self._colour_for_util(gpu_d['power_pct'])),  f" {gpu_d['power_pct']:5.1f}%")
            line3 = Text.assemble(("MEM ", "dim"), Text(_bar(gpu_d['mem_pct'], bar_w),   style=self._colour_for_util(gpu_d['mem_pct'])),    f" {gpu_d['mem_pct']:5.1f}%")
            gpu_panels.append(Panel(Group(line1, line2, line3), title=f"GPU{gpu_d['id']} - {role}", box=box.SQUARE, padding=(0, 1)))
        
        # build a 2-column grid (rows = ceil(N/2))
        tbl = Table.grid(expand=True, padding=0); tbl.add_column(ratio=1); tbl.add_column(ratio=1)
        for a, b in zip_longest(gpu_panels[0::2], gpu_panels[1::2], fillvalue=Panel("\n\n")): tbl.add_row(a, b) # filler for odd counts
        return Panel(tbl, title="GPU Performance", box=box.DOUBLE)

    def _base_stats(self) -> Panel:
        if not self._tracker_stats: return Panel(Text("waiting …"), title="Collection Stats.", box=box.SQUARE)
        try:
            self._hist["format_success"].append(self._tracker_stats["Format Success Rate - correct_answer_format"]*100); self._hist["inv_move_rate"].append(self._tracker_stats["Format Success Rate - invalid_move"]*100)
            self._hist["game_len"].append(self._tracker_stats["Game Length"]); self._hist["buffer_size"].append(self._buffer_size); bar_w = max(10, int(self.console.size.width * 0.48 - 35))
            line1 = Text.assemble(("Format Success: ", "dim"), Text(_scaled_hist(self._hist["format_success"], bar_w),  style="dim"), f" {self._hist['format_success'][-1]:5.2f}%")
            line2 = Text.assemble(("Inv. Move Rate: ", "dim"), Text(_scaled_hist(self._hist["inv_move_rate"], bar_w),   style="dim"), f" {self._hist['inv_move_rate'][-1]:5.2f}%")
            line3 = Text.assemble(("Game Length:    ", "dim"), Text(_scaled_hist(self._hist["game_len"], bar_w),        style="dim"), f" {self._hist['game_len'][-1]:5.2f} turns")
            line4 = Text.assemble(("Buffer Size:    ", "dim"), Text(_scaled_hist(self._hist["buffer_size"], bar_w),     style="dim"), f" {self._hist['buffer_size'][-1]:5.0f} samples")
            return Panel(Group(line1, line2, line3, line4), title=f"Collection Stats.", box=box.SQUARE, padding=(0, 1))
        except Exception as exc:
            return Panel(Text(f"Exception: {exc}"), title="Collection Stats.", box=box.SQUARE)
    
    def _ts(self) -> Panel:
        if not self._tracker_stats or not self._tracker_stats.get("TS"): return Panel(Text("waiting …"), title="TrueSkill", box=box.DOUBLE)
        bar_field = max(10, int(self.console.size.width * 0.45)-50); max_rows = max(3, int(self.console.size.height * 0.35))
        entries = [(uid, v["rating"].mu, v["rating"].sigma) for uid, v in self._tracker_stats["TS"].items()]
        entries.sort(key=lambda x: x[1], reverse=True); entries = entries[:max_rows]
        self._ts_idx: dict[str, int] = {}   # reset each draw
        for idx, (uid, _, _) in enumerate(entries): self._ts_idx[uid] = idx
        max_mu = max(mu for _, mu, _ in entries); bar_lines = []
        for uid, mu, sigma in entries:
            bar = "█" * int((mu/max_mu if max_mu else 0.0) * bar_field); bar_text = f"{bar:<{bar_field}}"  # pad to fixed width
            bar_lines.append(Text.assemble((f"[{self._ts_idx[uid]:<3}] {_trim_uid(uid):<25}", "bold"), "  ", Text(bar_text, style="cyan"), f"  μ {mu:6.2f}  σ {sigma:4.2f}"))
        title = f"TrueSkill (μ, σ)  – {len(self._tracker_stats['TS'])} models"
        return Panel(Group(*bar_lines), title=title, box=box.SQUARE, padding=(0, 1))

    def _heatmap(self) -> Panel: 
        if not self._tracker_stats or not self._tracker_stats.get("match_counts") or not getattr(self, "_ts_idx", None): return Panel(Text("waiting …"), title="Match Frequencies", box=box.SQUARE)
        uid_by_row = sorted(self._ts_idx, key=self._ts_idx.__getitem__); max_rows = max(3, int(self.console.size.height * 0.35)); uid_by_row = uid_by_row[:max_rows]
        def _cnt(a: str, b: str) -> int: return self._tracker_stats["match_counts"].get(tuple(sorted((a, b))), 0)
        max_cnt = max((_cnt(a, b) for a in uid_by_row for b in uid_by_row), default=1); light_rgb = (0, 43, 54); dark_rgb  = (255,  87,  34) # TODO somehow set light_rgb to be the users terminal color
        def _bg_style(pct: float) -> Style: return Style(bgcolor=Color.from_rgb(int(light_rgb[0]+pct*(dark_rgb[0]-light_rgb[0])), int(light_rgb[1]+pct*(dark_rgb[1]-light_rgb[1])), int(light_rgb[2]+pct*(dark_rgb[2]-light_rgb[2]))))
        tbl = Table.grid(padding=(0, 1)); tbl.add_row(""); header = [""] + [Text(f"{self._ts_idx[u]:>2}", style="bold") for u in uid_by_row]; tbl.add_row(*header)
        for ua in uid_by_row:
            row_cells = [Text(f"{self._ts_idx[ua]:>2}", style="bold")]
            for ub in uid_by_row:
                c = _cnt(ua, ub); row_cells.append(Text(f"{c:3}", style=_bg_style(c / max_cnt if max_cnt else 0.0)))
            tbl.add_row(*row_cells)
        return Panel(tbl, title="Match Counts", box=box.SQUARE)

    def _exploration(self) -> Panel: return Panel(Text("Not Implemented"), title="Exploration", box=box.DOUBLE) 
    async def run(self):
        layout = Layout(); layout.split_column(Layout(name="grid"), Layout(name="gpu", size=5*((self.gpu_count+self.gpu_count%2)//2)+2))
        layout["grid"].split_row(Layout(name="col1"), Layout(name="col2")); layout["col1"].split_column(Layout(name="exploration",ratio=1), Layout(name="heatmap", ratio=2))
        layout["col2"].split_column(Layout(name="ts"), Layout(name="bs", size=6)); asyncio.create_task(self._fetch_loop())  # Start background fetcher
        with Live(layout, console=self.console, auto_refresh=False) as live:
            while True:
                layout["gpu"].update(self._gpu_panel); layout["bs"].update(self._base_stats_panel); layout["ts"].update(self._ts_panel)
                layout["heatmap"].update(self._heatmap_panel); layout["exploration"].update(self._exploration_panel)
                live.refresh()
                await asyncio.sleep(2)

if __name__ == "__main__":
    ray.init(address="auto", namespace="unstable")   # connect to existing cluster
    term = TerminalInterface(tracker=ray.get_actor("Tracker"), buffer=ray.get_actor("Buffer"))
    asyncio.run(term.run())




import ray, torch, time, pathlib
from typing import List, Dict, Any, Optional

from unstable.buffers import BaseBuffer
from unstable.trackers import BaseTracker
from unstable.learners.utils import build_peft_model, enable_full_activation_ckpt
from unstable.utils import setup_logger


class BaseLearner:
    def __init__(self, model_name: str, lora_cfg: Dict[str,Any], batch_size: int, mini_batch_size: int, learning_rate: float, grad_clip: float, buffer: BaseBuffer, tracker: BaseTracker, model_registry, activation_checkpointing: bool=True, gradient_checkpointing: bool=True, use_trainer_cache: bool=False, initial_lora_path: Optional[str]=None): 
        # basically build the policy model and optimizer for policy model
        self.model_name, self.lora_cfg = model_name, lora_cfg
        self.buffer, self.tracker, self.model_registry = buffer, tracker, model_registry
        self.logger = setup_logger("learner", ray.get(tracker.get_log_dir.remote()))
        self.use_trainer_cache, self.gradient_checkpointing, self.activation_checkpointing = use_trainer_cache, gradient_checkpointing, activation_checkpointing
        self.batch_size, self.mini_batch_size, self.lr, self.grad_clip = batch_size, mini_batch_size, learning_rate, grad_clip
        self.gradient_acc_steps = self.batch_size // self.mini_batch_size # TODO maybe assert that divisible 
        self.ckpt_dir = pathlib.Path(ray.get(self.tracker.get_checkpoints_dir.remote())); self.ckpt_dir.mkdir(parents=True, exist_ok=True) # create ckpt dir

        torch.set_float32_matmul_precision('high')
        torch.set_default_dtype(torch.bfloat16)

        gpu_ids = ray.get_gpu_ids()
        self.device = (torch.device(f"cuda:{gpu_ids[0]}") if gpu_ids else torch.device("cpu"))
        self.policy_model, self.tokenizer = build_peft_model(model_name, self.device, lora_cfg, initial_lora_path)
        self.policy_model.to(torch.bfloat16)

        if not self.use_trainer_cache:      self.policy_model.config.use_cache = False
        if self.gradient_checkpointing:     self.policy_model.gradient_checkpointing_enable() # gradient checkpointing
        if self.activation_checkpointing:   enable_full_activation_ckpt(self.policy_model)       # activation checkpointing. Affords most of the vRAM savings
        
        self.policy_optimizer = torch.optim.AdamW(filter(lambda p: p.requires_grad, self.policy_model.parameters()), lr=learning_rate)
        self._step = 1; self._samples_seen = 0 # training counters

    def initialize_algorithm(self, cfg):    raise NotImplementedError
    def _update(self, batch):               raise NotImplementedError
    def train(self, iterations: int):
        self.logger.info("Starting training loop")

        while self._step < iterations:
            try:
                while (ray.get(self.buffer.size.remote()) < self.batch_size * 1.5): time.sleep(0.2) # wait until enough data is available
                self.logger.info("Enough data, starting learning step")
                batch: List = ray.get(self.buffer.get_batch.remote(self.batch_size)); self._samples_seen += self.batch_size
                accumulated_metrics = self._update(batch=batch) # handled by specific algo implementations

                log = {f"{k}": v for k, v in accumulated_metrics.items()}
                log.update({"step": self._step,  "samples_seen": self._samples_seen,  "lr": self.policy_optimizer.param_groups[0]["lr"], "policy_grad_norm": sum(p.grad.data.norm(2).item()**2 for p in self.policy_model.parameters() if p.grad is not None) ** 0.5})
                self.tracker.log_learner.remote(log)

                # save & register the updated checkpoint
                ckpt_path = self._save_checkpoint()
                try:
                    self.model_registry.add_checkpoint.remote(uid=f"ckpt-{self._step}", path=ckpt_path, iteration=self._step)
                    self.logger.info(f"Registered new ckpt: {ckpt_path}, ckpt-{self._step}")
                except Exception as exc: self.logger.info(f"Exception when adding checkpoint: {exc}")
                self.logger.info(f"registered new ckpt -> {ckpt_path} for iteration{self._step}")
                self._step += 1
            except Exception as exc: self.logger.info(f"Exception in learner loop: {exc}")

        self.logger.info("[Learner] training finished.")
        self.buffer.stop.remote()

    def _save_checkpoint(self):
        ckpt_dir = self.ckpt_dir / f"iteration-{self._step}"
        ckpt_dir.mkdir(parents=True, exist_ok=True)
        self.policy_model.save_pretrained(ckpt_dir, save_adapter=True)
        return ckpt_dir




import os, re, ray, time, wandb, collections, datetime, logging, numpy as np
from typing import Optional, Union, Dict
from unstable.utils import setup_logger

from unstable._types import PlayerTrajectory, GameInformation
from unstable.utils import write_game_information_to_file
Scalar = Union[int, float, bool]

class BaseTracker:
    def __init__(self, run_name: str):
        self.run_name = run_name 
        self._build_output_dir()

    def _build_output_dir(self):
        self.output_dir = os.path.join("outputs", str(datetime.datetime.now().strftime('%Y-%m-%d')), str(datetime.datetime.now().strftime('%H-%M-%S')), self.run_name)
        os.makedirs(self.output_dir)
        self.output_dirs = {}
        for folder_name in ["training_data", "eval_data", "checkpoints", "logs"]: 
            self.output_dirs[folder_name] =  os.path.join(self.output_dir, folder_name); os.makedirs(self.output_dirs[folder_name], exist_ok=True)

    def get_checkpoints_dir(self):  return self.output_dirs["checkpoints"]
    def get_train_dir(self):        return self.output_dirs["training_data"]
    def get_eval_dir(self):         return self.output_dirs["eval_data"]
    def get_log_dir(self):          return self.output_dirs["logs"]
    def add_trajectory(self, trajectory: PlayerTrajectory, env_id: str): raise NotImplementedError
    def add_eval_episode(self, episode_info: Dict, final_reward: int, player_id: int, env_id: str, iteration: int): raise NotImplementedError
    def log_lerner(self, info_dict: Dict): raise NotImplementedError

    
@ray.remote
class Tracker(BaseTracker): 
    FLUSH_EVERY = 64
    def __init__(self, run_name: str, wandb_project: Optional[str]=None):
        super().__init__(run_name=run_name)
        self.logger = setup_logger("tracker", self.get_log_dir())
        self.use_wandb = False
        if wandb_project: wandb.init(project=wandb_project, name=run_name); self.use_wandb = True; wandb.define_metric("*", step_metric="learner/step")
        self._m: Dict[str, collections.deque] = collections.defaultdict(lambda: collections.deque(maxlen=512))
        self._buffer: Dict[str, Scalar] = {}
        self._n = {}
        self._last_flush = time.monotonic()
        self._interface_stats = {"gpu_tok_s": {}, "TS": {}, "exploration": {}, "match_counts": {}, "format_success": None, "inv_move_rate": None, "game_len": None}

    def _put(self, k: str, v: Scalar): self._m[k].append(v)
    def _agg(self, p: str) -> dict[str, Scalar]: return {k: float(np.mean(dq)) for k, dq in self._m.items() if k.startswith(p)}
    def _flush_if_due(self):
        if time.monotonic()-self._last_flush >= self.FLUSH_EVERY:
            if self._buffer and self.use_wandb:
                try: wandb.log(self._buffer)
                except Exception as e: self.logger.warning(f"wandb.log failed: {e}")
            self._buffer.clear(); self._last_flush=time.monotonic()

    def add_player_trajectory(self, traj: PlayerTrajectory, env_id: str):
        try:
            reward = traj.final_reward; player_id = traj.pid
            self._put(f"collection-{env_id}/reward", reward)
            self._put(f"collection-{env_id}/Win Rate", int(reward>0))
            self._put(f"collection-{env_id}/Loss Rate", int(reward<0))
            self._put(f"collection-{env_id}/Draw", int(reward==0))
            self._put(f"collection-{env_id}/Reward (pid={traj.pid})", reward)
            self._put(f"collection-{env_id}/Game Length", traj.num_turns)
            for idx in range(len(traj.obs)):
                self._put(f"collection-{env_id}/Respone Length (char)", len(traj.actions[idx]))
                self._put(f"collection-{env_id}/Observation Length (char)", len(traj.obs[idx]))
                for k, v in traj.format_feedbacks[idx].items(): self._put(f"collection-{env_id}/Format Success Rate - {k}", v)
            self._n[f"collection-{env_id}"] = self._n.get(f"collection-{env_id}", 0) + 1
            self._put(f"collection-{env_id}/step", self._n[f"collection-{env_id}"])
            self._buffer.update(self._agg('collection-')); self._flush_if_due()
        except Exception as exc:
            self.logger.info(f"Exception when adding trajectory to tracker: {exc}")

    def add_eval_game_information(self, game_information: GameInformation, env_id: str):
        try:
            eval_reward = game_information.final_rewards.get(game_information.eval_model_pid, 0.0)
            _prefix = f"evaluation-{env_id}" if not game_information.eval_opponent_name else f"evaluation-{env_id} ({game_information.eval_opponent_name})"
            self._put(f"{_prefix}/Reward", eval_reward)
            self._put(f"{_prefix}/Reward (pid={game_information.eval_model_pid})", eval_reward)
            self._put(f"{_prefix}/Win Rate",  int(eval_reward>0))
            self._put(f"{_prefix}/Loss Rate", int(eval_reward<0))
            self._put(f"{_prefix}/Draw Rate", int(eval_reward==0))
            self._n[_prefix] = self._n.get(_prefix, 0) + 1
            self._put(f"{_prefix}/step", self._n[_prefix])
            self._buffer.update(self._agg('evaluation-')); self._flush_if_due()

            # try storing the eval info to file
            write_game_information_to_file(game_info=game_information, filename=os.path.join(self.get_eval_dir(), f"{env_id}-{game_information.game_idx}.csv"))

        except Exception as exc:
            self.logger.info(f"Exception when adding game_info to tracker: {exc}")

    def log_model_registry(self, ts_dict: dict[str, dict[str, float]], match_counts: dict[tuple[str, str], int]):
        self._interface_stats.update({"TS": ts_dict, "exploration": None, "match_counts": match_counts})

    def log_inference(self, actor: str, gpu_ids: list[int], stats: dict[str, float]):
        for key in stats: self._put(f"inference/{actor}/{key}", stats[key])
        for gpu_id in gpu_ids: self._interface_stats["gpu_tok_s"][gpu_id] = stats["tok_s"]
        self._buffer.update(self._agg('inference'))
    
    def log_learner(self, info: dict):
        self._m.update({f"learner/{k}": v for k, v in info.items()})
        self._m.update()
        self._buffer.update(self._agg("learner")); self._flush_if_due()

    def get_interface_info(self): 
        for inf_key in ["Game Length", "Format Success Rate - correct_answer_format", "Format Success Rate - invalid_move"]: 
            self._interface_stats[inf_key] = np.mean([float(np.mean(dq)) for k,dq in self._m.items() if inf_key in k])
        return self._interface_stats



from unstable.reward_transformations.transformation_final import ComposeFinalRewardTransforms, RoleAdvantageFormatter, RoleAdvantageByEnvFormatter
from unstable.reward_transformations.transformation_step import ComposeStepRewardTransforms, RewardForFormat, PenaltyForInvalidMove
from unstable.reward_transformations.transformation_sampling import ComposeSamplingRewardTransforms, NormalizeRewards, NormalizeRewardsByEnv
__all__ = ["ComposeFinalRewardTransforms", "RoleAdvantageFormatter", "RoleAdvantageByEnvFormatter", "ComposeStepRewardTransforms", "RewardForFormat", "PenaltyForInvalidMove", "ComposeSamplingRewardTransforms", "NormalizeRewards", "NormalizeRewardsByEnv"]


import ray, time
from typing import List, Sequence, Optional

import unstable
import unstable.reward_transformations as retra


_DEFAULT_LORA_CFG = {"lora_rank": 32, "lora_alpha": 32, "lora_dropout": 0.0, "target_modules": ["q_proj","k_proj","v_proj","o_proj","gate_proj", "up_proj","down_proj"]}
_ENV_SAMPLERS = {"random": unstable.samplers.env_samplers.UniformRandomEnvSampler}
_OPP_SAMPLERS = {"none": unstable.samplers.model_samplers.BaseModelSampler, "mirror": unstable.samplers.model_samplers.BaseModelSampler, "fixed": unstable.samplers.model_samplers.FixedOpponentModelSampler}
_STEP_BUFFER_ALGOS = ["reinforce"]
_EPISODE_BUFFER_ALGOS = ["a2c"]
_ALGOS = {"reinforce": unstable.REINFORCELearner, "a2c": unstable.A2CLearner}
def _default_vllm_cfg(model_name: str, lora_cfg: dict, max_generation_len: int) -> dict: return {"model_name": model_name, "temperature": 0.6, "max_tokens": max_generation_len, "max_parallel_seq": 128, "max_loras": 8, "lora_config": lora_cfg, "max_model_len": 8192}

class _UBRun:
    def __init__(self, *, collector, learner): self.collector, self.learner = collector, learner
    def start(self, learning_steps: int = 200, num_collection_workers: int = 256, num_eval_workers: int = 16):
        try:
            self.collector.collect.remote(num_train_workers=num_collection_workers, num_eval_workers=num_eval_workers)
            ray.get(self.learner.train.remote(learning_steps))
        finally:
            ray.kill(collector, no_restart=True); ray.shutdown()

def build(*, model_name: str, train_envs: Sequence[unstable.TrainEnvSpec], eval_envs: Optional[Sequence[unstable.EvalEnvSpec]]=None, env_sampling_strategy: str = "random", opponent_sampling_strategy: str = "none", fixed_opponents: Sequence[str] = ["google/gemini-2.0-flash-lite-001"], algorithm: str = "reinforce", max_train_len: Optional[int]=None, max_generation_len: int=4096, batch_size: int=384, mini_batch_size: int=1, learning_rate: float=1e-5, gradient_clipping: float=0.2, activation_checkpointing: bool=True, gradient_checkpointing: bool=True, use_trainer_cache: bool = False, buffer_size: Optional[int]=None, lora_config: Optional[dict]=None, vllm_config: Optional[dict]=None, wandb_project: str="UnstableBaselines"):
    # Ray init
    ray.init(namespace="unstable")  
    
    # env sampler
    assert env_sampling_strategy in _ENV_SAMPLERS, f"env_sampling_strategy='{env_sampling_strategy}' not found. Please use one of: {list(_ENV_SAMPLERS.keys())}"
    env_sampler = _ENV_SAMPLERS[env_sampling_strategy](train_env_specs=train_envs, eval_env_specs=eval_envs)

    # tracker
    tracker = unstable.Tracker.options(name="Tracker").remote(run_name=f"UnstableBaselines-{model_name.split('/')[-1]}-{env_sampler.env_list()}-{int(time.time())}", wandb_project="UnstableBaselines") 

    # initialize model registry
    model_registry = unstable.ModelRegistry.options(name="ModelRegistry").remote(tracker=tracker)
    ray.get(model_registry.add_checkpoint.remote(uid="base", path=None, iteration=0))
    for f_opp_name in fixed_opponents: ray.get(model_registry.add_fixed.remote(name=f_opp_name))

    # initialize opponent sampler
    assert opponent_sampling_strategy in _OPP_SAMPLERS, f"opponent_sampling_strategy='{opponent_sampling_strategy}' not found. Please use one of: {list(_OPP_SAMPLERS.keys())}"
    model_sampler = _OPP_SAMPLERS[opponent_sampling_strategy](model_registry=model_registry)

    # build game scheduler
    game_scheduler = unstable.GameScheduler.options(name="GameScheduler").remote(model_sampler=model_sampler, env_sampler=env_sampler, logging_dir=ray.get(tracker.get_log_dir.remote()))

    # build buffer TODO maybe move the reward transformations outside
    buffer_size = buffer_size or batch_size*2
    if algorithm in _STEP_BUFFER_ALGOS: buffer = unstable.StepBuffer.options(name="Buffer").remote(max_buffer_size=buffer_size, tracker=tracker, final_reward_transformation=retra.ComposeFinalRewardTransforms([retra.RoleAdvantageByEnvFormatter()]), step_reward_transformation=retra.ComposeStepRewardTransforms([retra.RewardForFormat(1.5), retra.PenaltyForInvalidMove(1.0, -1.0)]), sampling_reward_transformation=retra.ComposeSamplingRewardTransforms([retra.NormalizeRewardsByEnv(True)]))
    elif algorithm in _EPISODE_BUFFER_ALGOS: buffer = unstable.EpisodeBuffer.options(name="Buffer").remote(max_buffer_size=buffer_size, tracker=tracker, final_reward_transformation=retra.ComposeFinalRewardTransforms([retra.RoleAdvantageByEnvFormatter()]), step_reward_transformation=retra.ComposeStepRewardTransforms([retra.RewardForFormat(1.5), retra.PenaltyForInvalidMove(1.0, -1.0)]), sampling_reward_transformation=retra.ComposeSamplingRewardTransforms([retra.NormalizeRewardsByEnv(True)]))
    else: raise NotImplementedError(f"The algorithm used ({algorithm}) has not been allocated to a specific buffer type.")

    # initialize the collector
    _lora_cfg = lora_config or _DEFAULT_LORA_CFG
    collector = unstable.Collector.options(name="Collector").remote(vllm_config=vllm_config or _default_vllm_cfg(model_name, _lora_cfg, max_generation_len), tracker=tracker, buffer=buffer, game_scheduler=game_scheduler)

    # initialize the learner
    assert algorithm in _ALGOS, f"algorithm='{algorithm}' not found. Please use one of: {list(_ALGOS.keys())}"
    learner = _ALGOS[algorithm].options(num_gpus=1, name="Learner").remote(model_name=model_name, lora_cfg=_lora_cfg, batch_size=batch_size, mini_batch_size=mini_batch_size, learning_rate=learning_rate, grad_clip=gradient_clipping, buffer=buffer, tracker=tracker, model_registry=model_registry, activation_checkpointing=activation_checkpointing, gradient_checkpointing=gradient_checkpointing, use_trainer_cache=use_trainer_cache)
    match algorithm:
        case "reinforce":   ray.get(learner.initialize_algorithm.remote(max_train_len=max_train_len, max_generation_len=max_generation_len))
        case "a2c":         ray.get(learner.initialize_algorithm.remote(infer_mini_batch_size=16, critic_learning_rate=5e-5, normalize_adv=True, max_train_len=max_train_len, max_generation_len=max_generation_len)) # TODO find better solution
        case _:             ray.get(learner.initialize_algorithm.remote())

    return _UBRun(collector=collector, learner=learner)
